



test.pyutilib --help
IF %ERRORLEVEL% NEQ 0 exit 1
pyutilib_test_driver --help
IF %ERRORLEVEL% NEQ 0 exit 1
where lbin
IF %ERRORLEVEL% NEQ 0 exit 1
where lpython
IF %ERRORLEVEL% NEQ 0 exit 1
dispatch_srvr --help
IF %ERRORLEVEL% NEQ 0 exit 1
exit 0
